"""File-related utility functions.

Implements: `load_file`, `dump_file`, `load_json_file`,
`dump_json_file`, `get_parent_dir_path`, `get_dir_checksum`,
`get_file_checksum`, `rel_to_abs_path`, `read_last_n_lines`,
`expect_file_contents`, `render_directory_tree`, `list_directory`"""

from __future__ import annotations
from typing import Any, Optional
import fnmatch
import re
import traceback
import hashlib
import json
import os
import tum_esm_utils


def load_file(path: str) -> str:
    """Load the content of a file."""
    with open(path, "r") as f:
        return f.read()


def dump_file(path: str, content: str) -> None:
    """Dump content to a file."""
    with open(path, "w") as f:
        f.write(content)


def load_binary_file(path: str) -> bytes:
    """Load binary content of a file."""
    with open(path, "rb") as f:
        return f.read()


def dump_binary_file(path: str, content: bytes) -> None:
    """Dump binary content to a file."""
    with open(path, "wb") as f:
        f.write(content)


def load_json_file(path: str) -> Any:
    """Load the content of a JSON file."""
    with open(path, "r") as f:
        return json.load(f)


def dump_json_file(path: str, content: Any, indent: Optional[int] = 4) -> None:
    """Dump content to a JSON file."""
    with open(path, "w") as f:
        json.dump(content, f, indent=indent)


def get_parent_dir_path(script_path: str, current_depth: int = 1) -> str:
    """Get the absolute path of a parent directory based on the
    current script path. Simply pass the `__file__` variable of
    the current script to this function. Depth of 1 will return
    the direct parent directory of the current script."""

    assert current_depth > 0, "depth must be greater than 0"
    output = os.path.dirname(os.path.abspath(script_path))
    for _ in range(current_depth - 1):
        output = os.path.dirname(output)
    return output


def get_dir_checksum(path: str) -> str:
    """Get the checksum of a directory using md5deep."""

    assert os.path.isdir(path), f"{path} is not a directory"
    return tum_esm_utils.shell.run_shell_command(f"md5deep -r -b {path} | sort -u | md5sum")


def get_file_checksum(path: str) -> str:
    """Get the checksum of a file using MD5 from `haslib`.

    Significantly faster than `get_dir_checksum` since it does
    not spawn a new process."""

    assert os.path.isfile(path), f"{path} is not a file"
    with open(path, "rb") as f:
        return hashlib.md5(f.read()).hexdigest()


def rel_to_abs_path(*path: str) -> str:
    """Convert a path relative to the caller's file to an absolute path.

    Inside file `/home/somedir/somepath/somefile.py`, calling
    `rel_to_abs_path("..", "config", "config.json")` will
    return `/home/somedir/config/config.json`.

    Credits to https://stackoverflow.com/a/59004672/8255842"""

    return os.path.normpath(
        os.path.join(
            os.path.dirname(traceback.extract_stack()[-2].filename),
            *path,
        )
    )


def read_last_n_lines(
    file_path: str,
    n: int,
    ignore_trailing_whitespace: bool = False,
) -> list[str]:
    """Read the last `n` lines of a file.

    The function returns less than `n` lines if the file has less than `n` lines.
    The last element in the list is the last line of the file.

    This function uses seeking in order not to read the full file. The simple
    approach of reading the last 10 lines would be:

    ```python
    with open(path, "r") as f:
        return f.read().split("\\n")[:-10]
    ```

    However, this would read the full file and if we only need to read 10 lines
    out of a 2GB file, this would be a big waste of resources.

    The `ignore_trailing_whitespace` option to crop off trailing whitespace, i.e.
    only return the last `n` lines that are not empty or only contain whitespace."""

    with open(file_path, "rb") as f:
        f.seek(-1, os.SEEK_END)

        if ignore_trailing_whitespace:
            while f.read(1) in [b"\n", b" ", b"\t"]:
                try:
                    f.seek(-2, os.SEEK_CUR)
                except OSError:
                    # reached the beginning of the file
                    return [""]

            f.seek(-1, os.SEEK_CUR)
            # now the cursor is right before the last
            # character that is not a newline or a space

        last_characters: bytes = b""
        new_line_chars_seen: int = 0

        while True:
            try:
                new_character = f.read(1)
                if new_character == b"\n":
                    new_line_chars_seen += 1
                if new_line_chars_seen == n:
                    break
                last_characters += new_character
                f.seek(-2, os.SEEK_CUR)
            except OSError:
                # reached the beginning of the file
                break

    return last_characters.decode()[::-1].split("\n")


def expect_file_contents(
    filepath: str,
    required_content_blocks: list[str] = [],
    forbidden_content_blocks: list[str] = [],
) -> None:
    """Assert that the given file contains all of the required content
    blocks, and/or none of the forbidden content blocks.

    Args:
        filepath:                 The path to the file.
        required_content_blocks:  A list of strings that must be present in the file.
        forbidden_content_blocks: A list of strings that must not be present in the file.
    """

    with open(filepath, "r") as f:
        file_content = f.read()

    for b in required_content_blocks:
        assert b in file_content, f'required log content block not found "{b}"'

    for b in forbidden_content_blocks:
        assert b not in file_content, f'forbidden log content block found "{b}"'


def render_directory_tree(
    root: str,
    ignore: list[str] = [],
    max_depth: Optional[int] = None,
    root_alias: Optional[str] = None,
    directory_prefix: Optional[str] = "📁 ",
    file_prefix: Optional[str] = "📄 ",
) -> Optional[str]:
    """Render a file tree as a string.

    Example:

    ```
    📁 <config.general.data.results>
    ├─── 📁 bundle
    │    ├─── 📄 __init__.py
    │    ├─── 📄 load_results.py
    │    └─── 📄 main.py
    ├─── 📁 profiles
    │    ├─── 📄 __init__.py
    │    ├─── 📄 cache.py
    │    ├─── 📄 download_logic.py
    │    ├─── 📄 generate_queries.py
    │    ├─── 📄 main.py
    │    ├─── 📄 std_site_logic.py
    │    └─── 📄 upload_logic.py
    ├─── 📁 retrieval
    │    ├─── 📁 algorithms
    ...
    ```

    Args:
        root:              The root directory to render.
        ignore:            A list of patterns to ignore. If the basename of a directory
                           matches any of the patterns, the directory is ignored.
        max_depth:         The maximum depth to render. If `None`, render the full tree.
        root_alias:        An alias for the root directory. If `None`, the basename of
                           the root directory is used. In the example above, the root
                           directory is was aliased to `<config.general.data.results>`.
        directory_prefix:  The prefix to use for directories.
        file_prefix:       The prefix to use for files.

    Returns: The directory tree as a string. If the root directory is ignored, `None`.
    """

    if any([fnmatch.fnmatch(os.path.basename(root), pattern) for pattern in ignore]):
        return None

    root_name = os.path.basename(root) if root_alias is None else root_alias

    if directory_prefix is None:
        directory_prefix = ""
    if file_prefix is None:
        file_prefix = ""

    if os.path.isdir(root):
        out = f"{directory_prefix}{root_name}"
        if max_depth == 0:
            return out

        sublists: list[str] = []
        for item in sorted(os.listdir(root)):
            sd = render_directory_tree(
                os.path.join(root, item),
                ignore=ignore,
                max_depth=(max_depth - 1) if max_depth is not None else None,
            )
            if sd is not None:
                sublists.append(sd)

        sublists = sorted(sublists)
        for i, item in enumerate(sublists):
            for j, line in enumerate(item.split("\n")):
                prefix: str
                if i < (len(sublists) - 1):
                    if j == 0:
                        prefix = "├───"
                    else:
                        prefix = "│   "
                else:
                    if j == 0:
                        prefix = "└───"
                    else:
                        prefix = "    "
                out += f"\n{prefix} {line}"
        return out
    if os.path.isfile(root):
        return f"{file_prefix}{root_name}"
    return ""


def list_directory(
    path: str,
    regex: Optional[str] = None,
    ignore: Optional[list[str]] = None,
    include_directories: bool = True,
    include_files: bool = True,
    include_links: bool = True,
) -> list[str]:
    """List the contents of a directory based on certain criteria. Like `os.listdir`
    with superpowers. You can filter the list by a regex or you can ignore Unix shell
    style patterns like `*.lock`.

    Args:
        path:                 The path to the directory.
        regex:                A regex pattern to match the item names against.
        ignore:               A list of patterns to ignore. If the basename of an item
                              matches any of the patterns, the item is ignored.
        include_directories:  Whether to include directories in the output.
        include_files:        Whether to include files in the output.
        include_links:        Whether to include symbolic links in the output.

    Returns: A list of items in the directory that match the criteria.
    """

    assert os.path.isdir(path), f"Path {path} is not a directory"
    files = os.listdir(path)
    if regex is not None:
        files = [f for f in files if re.match(regex, f)]
    if ignore is not None:
        files = [f for f in files if not any([fnmatch.fnmatch(f, pattern) for pattern in ignore])]
    if not include_directories:
        files = [f for f in files if not os.path.isdir(os.path.join(path, f))]
    if not include_files:
        files = [f for f in files if not os.path.isfile(os.path.join(path, f))]
    if not include_links:
        files = [f for f in files if not os.path.islink(os.path.join(path, f))]

    return files



"""Functions to read NCEP profiles."""

import polars as pl


# fmt: off
map_columns = [
    "Height", "Temp", "Pressure", "Density", "h2o", "hdo", "co2", "n2o", "co", "ch4", "hf", "o2",
    "gravity",
]
mod_columns = [
    "Pressure", "Temperature", "Height", "MMW", "H2O", "RH", "EPV", "PT", "EqL", "O3", "CO",
]
vmr_columns = [
    "Altitude", "H2O", "CO2", "O3", "N2O", "CO", "CH4", "O2", "NO", "SO2", "NO2", "NH3", "HNO3",
    "OH", "HF", "HCl", "HBr", "HI", "ClO", "OCS", "H2CO", "HOCl", "HO2", "H2O2", "HONO", "HO2NO2",
    "N2O5", "ClNO3", "HCN", "CH3F", "CH3Cl", "CF4", "CCl2F2", "CCl3F", "CH3CCl3", "CCl4", "COF2",
    "COFCl", "C2H6", "C2H4", "C2H2", "N2", "CHClF2", "COCl2", "CH3Br", "CH3I", "HCOOH", "H2S",
    "CHCl2F", "HDO", "SF6", "F113", "ClCN", "F142b", "dust_m", "PH3", "CH3OH", "CH3SH", "CH3CHO",
    "CH3CN", "PAN", "NF3", "ClOOCl", "ClClO2", "ClOClO", "CHF3", "f141b", "CH3COOH", "cirrus6",
    "cirrus15", "C3H8", "D2O", "sa_venus", "C6H6", "C3H6", "CH3COCH3", "CFH2CF3", "n-C4H10",
    "C5H8", "LUFT",
]
# fmt: on

mod_column_widths = [10, 14, 11, 11, 14, 10, 14, 12, 11, 13, 13]
mod_column_offsets = [0, *([sum(mod_column_widths[:i]) for i in range(1, len(mod_columns))])]
vmr_column_widths = [9, *([11] * (len(vmr_columns) - 2)), 10]
vmr_column_offsets = [0, *([sum(vmr_column_widths[:i]) for i in range(1, len(vmr_columns))])]


def load_ggg2020_map(filepath: str) -> pl.DataFrame:
    """Load the Atmospheric profile from a GGG2020 map file."""
    return (
        pl.read_csv(
            filepath,
            has_header=False,
            skip_rows=12,
            infer_schema=False,
            separator="v",
            schema={"full_str": pl.Utf8},
            n_threads=1,
        )
        .lazy()
        .with_columns(pl.col("full_str").str.split(",").alias("column_list"))
        .drop("full_str")
        .with_columns(
            *[
                pl.col("column_list")
                .list.get(i)
                .str.strip_chars(" ")
                .cast(pl.Float64)
                .alias(column)
                for i, column in enumerate(map_columns)
            ]
        )
        .drop("column_list")
        .collect()
    )


def load_ggg2020_mod(filepath: str) -> pl.DataFrame:
    """Load the Atmospheric profile from a GGG2020 mod file."""

    return (
        pl.read_csv(
            filepath,
            has_header=False,
            skip_rows=7,
            infer_schema=False,
            schema={"full_str": pl.Utf8},
            n_threads=1,
        )
        .lazy()
        .with_columns(
            [
                pl.col("full_str")
                .str.slice(offset, width)
                .str.strip_chars()
                .cast(pl.Float64)
                .alias(column)
                for column, offset, width in zip(mod_columns, mod_column_offsets, mod_column_widths)
            ]
        )
        .drop("full_str")
        .collect()
    )


def load_ggg2020_vmr(filepath: str) -> pl.DataFrame:
    """Load the Atmospheric profile from a GGG2020 vmr file."""
    return (
        pl.read_csv(
            filepath,
            has_header=False,
            skip_rows=8,
            infer_schema=False,
            schema={"full_str": pl.Utf8},
            n_threads=1,
        )
        .lazy()
        .with_columns(
            [
                pl.col("full_str")
                .str.slice(offset, width)
                .str.strip_chars()
                .cast(pl.Float64)
                .alias(column)
                for column, offset, width in zip(vmr_columns, vmr_column_offsets, vmr_column_widths)
            ]
        )
        .drop("full_str")
        .collect()
    )



"""Implements validator utils for use with pydantic models.

Implements: `StrictFilePath`, `StrictDirectoryPath`"""

from __future__ import annotations
from typing import Literal, Optional
import os
import pydantic


class StrictFilePath(pydantic.RootModel[str]):
    """A pydantic model that validates a file path.

    Example usage:

    ```python
    class MyModel(pyndatic.BaseModel):
        path: StrictFilePath

    m = MyModel(path='/path/to/file') # validates that the file exists
    ```

    The validation can be ignored by setting the context variable:

    ```python
    m = MyModel.model_validate(
        {"path": "somenonexistingpath"},
        context={"ignore-path-existence": True},
    ) # does not raise an error
    ```
    """

    root: str

    @pydantic.field_validator("root")
    @classmethod
    def path_should_exist(cls, v: str, info: pydantic.ValidationInfo) -> str:
        ignore_path_existence = (
            (info.context.get("ignore-path-existence") == True)
            if isinstance(info.context, dict)
            else False
        )
        if (not ignore_path_existence) and (not os.path.isfile(v)):
            raise ValueError("File does not exist")
        return v


class StrictDirectoryPath(pydantic.RootModel[str]):
    """A pydantic model that validates a directory path.

    Example usage:

    ```python
    class MyModel(pyndatic.BaseModel):
        path: StrictDirectoryPath

    m = MyModel(path='/path/to/directory') # validates that the directory exists
    ```

    The validation can be ignored by setting the context variable:

    ```python
    m = MyModel.model_validate(
        {"path": "somenonexistingpath"},
        context={"ignore-path-existence": True},
    ) # does not raise an error
    ```
    """

    root: str

    @pydantic.field_validator("root")
    @classmethod
    def path_should_exist(cls, v: str, info: pydantic.ValidationInfo) -> str:
        ignore_path_existence = (
            (info.context.get("ignore-path-existence") == True)
            if isinstance(info.context, dict)
            else False
        )
        if (not ignore_path_existence) and (not os.path.isdir(v)):
            raise ValueError("Directory does not exist")
        return v


class Version(pydantic.RootModel[str]):
    """A version string in the format of MAJOR.MINOR.PATCH[-(alpha|beta|rc).N]"""

    root: str = pydantic.Field(
        ...,
        pattern=r"^\d+\.\d+\.\d+(-(alpha|beta|rc)\.\d+)?$",
        examples=["1.2.3", "4.5.6-alpha.78", "7.8.9-beta.10", "11.12.13-rc.14"],
    )

    def as_tag(self) -> str:
        """Return the version string as a tag, i.e. vMAJOR.MINOR.PATCH..."""
        return "v" + self.root

    def as_identifier(self) -> str:
        """Return the version string as a number, i.e. MAJOR.MINOR.PATCH..."""
        return self.root

    def _split(self) -> tuple[int, int, int, Optional[tuple[Literal["alpha", "beta", "rc"], int]]]:
        """Split the version string into MAJOR, MINOR, PATCH, and TAG"""
        version, tag = self.root.split("-") if "-" in self.root else (self.root, "")
        major, minor, patch = map(int, version.split("."))
        if "-" in self.root:
            tags = tag.split(".")
            return major, minor, patch, (tags[0], int(tags[1]))  # type: ignore
        else:
            return major, minor, patch, None

    # add comparisons
    def __lt__(self, other: Version) -> bool:
        self_major, self_minor, self_patch, self_tag = self._split()
        other_major, other_minor, other_patch, other_tag = other._split()

        if self_major != other_major:
            return self_major < other_major
        if self_minor != other_minor:
            return self_minor < other_minor
        if self_patch != other_patch:
            return self_patch < other_patch

        if self_tag is None:
            return False

        if other_tag is None:
            return True

        assert (self_tag is not None) and (other_tag is not None)
        self_tag_type, self_tag_number = self_tag
        other_tag_type, other_tag_number = other_tag

        if (self_tag_type == "alpha") and (other_tag_type in ["beta", "rc"]):
            return True
        if (self_tag_type == "beta") and (other_tag_type == "rc"):
            return True
        if (self_tag_type == "beta") and (other_tag_type == "alpha"):
            return False
        if (self_tag_type == "rc") and (other_tag_type in ["alpha", "beta"]):
            return False

        assert self_tag_type == other_tag_type
        return self_tag_number < other_tag_number

    def __le__(self, other: Version) -> bool:
        return (self < other) or (self == other)

    def __gt__(self, other: Version) -> bool:
        return not (self <= other)

    def __ge__(self, other: Version) -> bool:
        return not (self < other)


class StricterBaseModel(pydantic.BaseModel):
    """The same as pydantic.BaseModel, but with stricter rules. It does not
    allow extra fields and validates assignments after initialization."""

    model_config = pydantic.ConfigDict(extra="forbid", validate_assignment=True)


_single_byte_as_dec_regex = r"((\d)|([1-9]\d)|(1\d\d)|(2[0-4]\d)|(25[0-5]))"


class StrictIPv4Adress(pydantic.RootModel[str]):
    """A pydantic model that validates an IPv4 address.

    Example usage:

    ```python
    class MyModel(pyndatic.BaseModel):
        ip: StrictIPv4Adress

    m = MyModel(ip='192.186.2.1')
    m = MyModel(ip='192.186.2.1:22')
    ```
    """

    root: str = pydantic.Field(
        ...,
        pattern="".join(
            [
                r"^",
                _single_byte_as_dec_regex,
                r"\.",
                _single_byte_as_dec_regex,
                r"\.",
                _single_byte_as_dec_regex,
                r"\.",
                _single_byte_as_dec_regex,
                r"(:\d{1,5})?",
                r"$",
            ]
        ),
    )



from typing import Any
import datetime
import skyfield.api
import tum_esm_utils


class Astronomy:
    """Astronomy utilities."""

    def __init__(self) -> None:
        """Initializes the Astronomy class, downloads the latest `de421.bsp` dataset."""

        self.planets: Any = skyfield.api.Loader(tum_esm_utils.files.rel_to_abs_path("."))(
            "de421.bsp"
        )
        self.timescale = skyfield.api.load.timescale()
        self.earth: Any = self.planets["Earth"]
        self.sun: Any = self.planets["Sun"]

    def get_sun_position(
        self,
        lat: float,
        lon: float,
        alt_asl: float,
        dt: datetime.datetime,
    ) -> tuple[float, float]:
        """Computes current sun elevation and azimuth in degrees."""

        skyfield_dt = self.timescale.from_datetime(dt)
        skyfield_location = self.earth + skyfield.api.wgs84.latlon(
            latitude_degrees=lat,
            longitude_degrees=lon,
            elevation_m=alt_asl,
        )
        altitude, azimuth, _ = (
            skyfield_location.at(skyfield_dt).observe(self.sun).apparent().altaz()
        )
        return float(altitude.degrees), float(azimuth.degrees)



"""Provides a HTTP interface to OPUS."""

from __future__ import annotations
from typing import Literal, Optional
import os
import time
import socket
import tenacity


class OpusHTTPInterface:
    """Interface to the OPUS HTTP interface.

    It uses the socket library, because the HTTP interface of OPUS does not
    reuturn valid HTTP/1 or HTTP/2 headers. It opens and closes a new socket
    because OPUS closes the socket after the answer has been sent.

    Raises:
        ConnectionError: If the connection to the OPUS HTTP interface fails or
                         if the response is invalid.
    """

    @staticmethod
    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(ConnectionError),
        reraise=True,
        stop=tenacity.stop_after_attempt(3),
        wait=tenacity.wait_fixed(5),
    )
    def request(
        request: str,
        timeout: float = 10.0,
        expect_ok: bool = False,
    ) -> list[str]:
        """Send a request to the OPUS HTTP interface and return the answer.

        Commands will be send to `GET http://localhost/OpusCommand.htm?<request>`.
        This function will retry the request up to 3 times and wait 5 seconds
        inbetween retries.

        Args:
            request:    The request to send.
            timeout:    The time to wait for the answer.
            expect_ok:  Whether the first line of the answer should be "OK".

        Returns:
            The answer lines.
        """

        return OpusHTTPInterface.request_without_retry(
            request, timeout=timeout, expect_ok=expect_ok
        )

    @staticmethod
    def request_without_retry(
        request: str,
        timeout: float = 10.0,
        expect_ok: bool = False,
    ) -> list[str]:
        """Send a request to the OPUS HTTP interface and return the answer.

        Commands will be send to `GET http://localhost/OpusCommand.htm?<request>`.

        Args:
            request:    The request to send.
            timeout:    The time to wait for the answer.
            expect_ok:  Whether the first line of the answer should be "OK".

        Returns:
            The answer lines.
        """

        answer_lines: Optional[list[str]] = None
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(timeout)
            s.connect(("localhost", 80))
            url = f"/OpusCommand.htm?{request.replace(' ', '%20')}"
            s.sendall(f"GET {url}\r\nHost: localhost\r\n\r\n".encode("utf-8"))
            answer = s.recv(4096).decode("utf-8").strip("\r\n\t ")
            answer_lines = [l.strip(" \r\t") for l in answer.split("\n")]
            answer_lines = [l for l in answer_lines if len(l) > 0]
            if expect_ok:
                assert len(answer_lines) >= 1
                assert answer_lines[0] == "OK"
            s.close()
            return answer_lines
        except:
            raise ConnectionError(
                f"Invalid response from OPUS HTTP interface: "
                + ("no answer" if answer_lines is None else str(answer_lines))
            )

    @staticmethod
    def get_version() -> str:
        """Get the version number, like `20190310`."""
        answer = OpusHTTPInterface.request("GET_VERSION")
        try:
            return answer[0]
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def get_version_extended() -> str:
        """Get the extended version number, like `8.2 Build: 8, 2, 28 20190310`."""
        answer = OpusHTTPInterface.request("GET_VERSION_EXTENDED")
        try:
            return answer[0]
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def is_working() -> bool:
        """Check if the OPUS HTTP interface is working. Does NOT raise a
        `ConnectionError` but only returns `True` or `False`."""
        try:
            answer = OpusHTTPInterface.request("COMMAND_SAY hello")
            assert answer[0] == "hello"
            return True
        except:
            return False

    @staticmethod
    def get_main_thread_id() -> int:
        """Get the process ID of the main thread of OPUS."""
        answer = OpusHTTPInterface.request("FIND_FUNCTION 0", expect_ok=True)
        try:
            return int(answer[1])
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def some_macro_is_running() -> bool:
        """Check if any macro is currently running.

        In theory, we could also check whether the correct macro is running using
        `READ_PARAMETER MPT` and `READ_PARAMETER MFN`. However, these variables do
        not seem to be updated right away, so we cannot rely on them."""

        main_thread_id = OpusHTTPInterface.get_main_thread_id()
        active_thread_ids: set[int] = set()

        # some common functions executed inside Macro routines that take some time
        common_functions = [
            "MeasureReference",
            "MeasureSample",
            "MeasureRepeated",
            "MeasureRapidTRS",
            "MeasureStepScanTrans",
            "UserDialog",
            "Baseline",
            "PeakPick",
            "Timer",
            "SendCommand",
        ]

        # check twice for any thread that is executing a common function
        for i in range(2):
            for function in common_functions:
                answer = OpusHTTPInterface.request(f"FIND_FUNCTION {function}")
                try:
                    if answer[0] == "OK":
                        for thread_id in answer[1:]:
                            active_thread_ids.add(int(thread_id))
                except:
                    raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")
            if i == 0:
                time.sleep(3)

        # the main thread always runs some common functions for some reason
        if main_thread_id in active_thread_ids:
            active_thread_ids.remove(main_thread_id)

        # if there is any thread that is not the main thread, then a macro is running
        return len(active_thread_ids) > 0

    @staticmethod
    def get_loaded_experiment() -> str:
        """Get the path to the currently loaded experiment."""
        OpusHTTPInterface.set_parameter_mode("opus")
        xpp_value = OpusHTTPInterface.read_parameter("XPP")
        exp_value = OpusHTTPInterface.read_parameter("EXP")
        return os.path.join(xpp_value, exp_value)

    @staticmethod
    def load_experiment(experiment_path: str) -> None:
        """Load an experiment file."""
        OpusHTTPInterface.request(f"LOAD_EXPERIMENT {experiment_path}", expect_ok=True)

    @staticmethod
    def start_macro(macro_path: str) -> int:
        """Start a macro. Returns the macro ID."""
        answer = OpusHTTPInterface.request(f"RUN_MACRO {macro_path}", expect_ok=True)
        try:
            return int(answer[1])
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def macro_is_running(macro_id: int) -> bool:
        """Check if the given macro is running. It runs `MACRO_RESULTS <macro_id>`
        under the hood. The OPUS documentation is ambiguous about the return value.
        It seems that 0 means "there is no result yet", i.e. the macro is still running"""
        answer = OpusHTTPInterface.request(f"MACRO_RESULTS {macro_id}", expect_ok=True)
        try:
            return int(answer[1]) == 0
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def stop_macro(macro_path_or_id: str | int) -> None:
        """Stop a macro given by its path or ID.

        Stopping a macro by its ID only works for our OPUS 8.X installations,
        but not our OPUS 7.X installations. Hence, it is recommended to always
        stop it by path."""

        if isinstance(macro_path_or_id, int):
            OpusHTTPInterface.request(f"KILL_MACRO {macro_path_or_id}", expect_ok=True)
        else:
            OpusHTTPInterface.request(
                f"KILL_MACRO {os.path.basename(macro_path_or_id)}", expect_ok=True
            )

    @staticmethod
    def unload_all_files() -> None:
        """Unload all files. This should be done before closing it."""
        OpusHTTPInterface.command_line("UnloadAll()")

    @staticmethod
    def close_opus() -> None:
        """Close OPUS."""
        OpusHTTPInterface.request("CLOSE_OPUS", expect_ok=True)

    @staticmethod
    def set_parameter_mode(variant: Literal["file", "opus"]) -> None:
        """Set the parameter mode to `FILE_PARAMETERS` or `OPUS_PARAMETERS`."""
        OpusHTTPInterface.request(f"{variant.upper()}_PARAMETERS", expect_ok=True)

    @staticmethod
    def read_parameter(parameter: str) -> str:
        """Read the value of a parameter."""
        answer = OpusHTTPInterface.request(f"READ_PARAMETER {parameter}", expect_ok=True)
        try:
            return answer[1]
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def write_parameter(parameter: str, value: str | int | float) -> None:
        """Update the value of a parameter."""
        OpusHTTPInterface.request(f"WRITE_PARAMETER {parameter} {value}", expect_ok=True)

    @staticmethod
    def get_language() -> str:
        """Get the current language."""
        answer = OpusHTTPInterface.request("GET_LANGUAGE", expect_ok=True)
        try:
            return answer[1]
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def get_username() -> str:
        """Get the current username."""
        answer = OpusHTTPInterface.request("GET_USERNAME", expect_ok=True)
        try:
            return answer[1]
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def get_path(literal: Literal["opus", "base", "data", "work"]) -> str:
        """Get the path to the given directory."""
        answer = OpusHTTPInterface.request(f"GET_{literal.upper()}PATH", expect_ok=True)
        try:
            return answer[1]
        except:
            raise ConnectionError(f"Invalid response from OPUS HTTP interface: {answer}")

    @staticmethod
    def set_processing_mode(mode: Literal["command", "execute", "request"]) -> None:
        """Set the processing mode to `COMMAND_MODE`, `EXECUTE_MODE`, or `REQUEST_MODE`."""
        OpusHTTPInterface.request(f"SET_{mode.upper()}_MODE", expect_ok=True)

    @staticmethod
    def command_line(command: str) -> Optional[str]:
        """Execute a command line command, i.e. `COMMAND_LINE <command>`."""
        answer = OpusHTTPInterface.request(f"COMMAND_LINE {command}", expect_ok=True)
        return None if len(answer) == 1 else answer[1]
